from collections import OrderedDict

from termcolor import cprint


def setup_yaml():
    """ https://stackoverflow.com/a/8661021 """
    import yaml

    def represent_dict_order(self, data):
        return self.represent_mapping('tag:yaml.org,2002:map', data.items())

    yaml.add_representer(OrderedDict, represent_dict_order)

    cprint('yaml is setup', 'green')


setup_yaml()


def show(obj):
    import yaml
    print(yaml.dump(obj))


def shows(obj):
    import yaml
    return yaml.dump(obj)



from hachiko.hachiko import AIOEventHandler, AIOWatchdog
from asyncio import coroutine, Queue, sleep
from sanic import response
from sanic.exceptions import RequestTimeout

from ml_dash.file_utils import path_match
from termcolor import cprint

from . import config
import json

subscriptions = []
watcher = None


class Handler(AIOEventHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @coroutine
    async def on_any_event(self, event):
        _event = dict(src_path=event.src_path, event_type=event.event_type, is_directory=event.is_directory)
        for que in subscriptions:
            await que.put(_event)
            # self._loop.create_task(que.put(event))


def setup_watch_queue(app, loop):
    print('setting up watch queue')
    start_watcher()
    cprint('watcher setup complete!', "green")


def start_watcher():
    global watcher

    handler = Handler()
    print('starting file watcher...')
    watcher = AIOWatchdog(config.Args.logdir, event_handler=handler)
    watcher.start()
    print('watcher start is complete')


import os


# server does not have access to a disconnect event.
# currently subscriptions only grows.
# Will add timeout based cleanup after.
async def file_events(request, file_path="", query="*"):
    q = Queue()
    subscriptions.append(q)

    async def streaming_fn(response):
        try:
            while True:
                print('subscription que started')
                file_event = await q.get()
                src_path = file_event['src_path']
                if src_path.startswith(os.path.join(config.Args.logdir, file_path)) and path_match(file_path, query):
                    file_event['src_path'] = src_path[len(config.Args.logdir):]
                print("=>>", file_event)
                response.write(f"data: {json.dumps(file_event)}\r\n\r\n".encode())
                sleep(0.1)
        # todo: this timeout doesn't really work.
        # todo: also add handling of stream is terminated logic (separate from above).
        except RequestTimeout:
            subscriptions.remove(q)

    return response.stream(streaming_fn, content_type="text/event-stream")
    # subscriptions.remove(q)



# SSE "protocol" is described here: http://mzl.la/UPFyxY
class ServerSentEvent(object):

    def __init__(self, data):
        self.data = data
        self.event = None
        self.id = None
        self.desc_map = {
            self.data: "data",
            self.event: "event",
            self.id: "id"
        }

    def __str__(self):
        if not self.data:
            return ""
        lines = [f"{v}: {k}" for k, v in self.desc_map.items() if k]
        return "%s\n\n" % "\n".join(lines)



from os.path import join, isabs

import numpy as np
import pandas as pd
from graphene import relay, ObjectType, String, List, ID, Int, Float
from graphene.types.generic import GenericScalar
from ml_dash.config import Args
from ml_dash.schema.files.file_helpers import read_dataframe


def get_column(df, key, stat_key):
    try:
        return df[key][stat_key].to_numpy().tolist()
    except:
        return []


def get_columns(df, keys, stat_key):
    return {k: get_column(df, k, stat_key) for k in keys}


class Series(ObjectType):
    class Meta:
        interfaces = relay.Node,

    path = String(description="the file path for the configuration file")

    prefix = String(description='path prefix for the metrics files')
    metrics_files = List(ID, description="List of metrics file IDs that we use to aggregate this series")

    _df = GenericScalar(description='the processed dataframe object that aggregates all metrics files.')

    window = Float(description="the window for the rolling average")

    label = String(description="the lable for the series")
    x_key = String(description="key for the x")
    y_key = String(description="key for the y axis")
    y_keys = List(String, description="list of keys for the y axis")

    # stype = SeriesTyes(description="the type of series data")

    # resolved from dataset
    x_data = GenericScalar(description="x data")
    y_mean = GenericScalar(description="y data from the mean of the window")
    # y_mode = GenericScalar(description="y data as from mode of the window")
    y_median = GenericScalar(description="y data as from mode of the window")
    y_min = GenericScalar(description="min in each bin")
    y_max = GenericScalar(description="max in each bin")
    y_25pc = GenericScalar(description="quarter quantile")
    y_75pc = GenericScalar(description="3/4th quantile")
    y_95pc = GenericScalar(description="95th quantile")
    y_05pc = GenericScalar(description="5th quantile")
    # Note: confidence level only applies to mean. So only group has it.
    # y_c95 = GenericScalar(description="95th confidence")
    y_count = GenericScalar(description="the number of datapoints used to compute each tick")

    warning = String(description="Warning Message")

    # todo: start time

    # todo: need to move the keys out, so that we can dropnan on the joint table.
    #   Otherwise the different data columns would not necessarily be the same length.
    def resolve_x_data(self, info):
        # note: new in 0.24.1.
        #  ~> df.value.dtype does NOT work for categorical data.
        _ = self._df['__x'].to_numpy()
        if np.issubdtype(_.dtype, np.datetime64):
            return (_.astype(int) / 1000).tolist()
        elif np.issubdtype(_.dtype, np.timedelta64):
            return (_.astype(int) / 1000).tolist()
        return _.tolist()

    def resolve_y_mean(self, info):
        if self.y_key is not None:
            return get_column(self._df, self.y_key, 'mean')
        return get_columns(self._df, self.y_keys, 'mean')

    # def resolve_y_mode(self, info):
    #     if self.y_key is not None:
    #         return self._df[self.y_key]['mode'].to_numpy().tolist()
    #     return {k: self._df[k]['mode'].to_numpy().tolist() for k in self.y_keys}

    def resolve_y_min(self, info):
        if self.y_key is not None:
            return get_column(self._df, self.y_key, 'min')
        return get_columns(self._df, self.y_keys, 'min')

    def resolve_y_max(self, info):
        if self.y_key is not None:
            return get_column(self._df, self.y_key, 'max')
        return get_columns(self._df, self.y_keys, 'max')

    def resolve_y_median(self, info):
        if self.y_key is not None:
            return get_column(self._df, self.y_key, '50%')
        return get_columns(self._df, self.y_keys, '50%')

    def resolve_y_25pc(self, info):
        if self.y_key is not None:
            return get_column(self._df, self.y_key, '25%')
        return get_columns(self._df, self.y_keys, '25%')

    def resolve_y_75pc(self, info):
        if self.y_key is not None:
            return get_column(self._df, self.y_key, '75%')
        return get_columns(self._df, self.y_keys, '75%')

    def resolve_y_95pc(self, info):
        if self.y_key is not None:
            return get_column(self._df, self.y_key, '95%')
        return get_columns(self._df, self.y_keys, '95%')

    def resolve_y_05pc(self, info):
        if self.y_key is not None:
            return get_column(self._df, self.y_key, '5%')
        return get_columns(self._df, self.y_keys, '5%')

    def resolve_y_count(self, info):
        if self.y_key is not None:
            return get_column(self._df, self.y_key, 'count')
        return get_columns(self._df, self.y_keys, 'count')

    @classmethod
    def get_node(cls, info, id):
        return Series(id)


def get_series(metrics_files=tuple(),
               prefix=None,
               head=None,
               tail=None,
               x_low=None,
               x_high=None,
               x_edge=None,  # OneOf('start', 'after', 'mid', 'mode')
               k=None,
               x_align=None,  # OneOf(int, 'left', 'right')
               x_key=None,
               y_key=None,
               y_keys=None,
               label=None):
    warning = None
    assert not y_key or not y_keys, "yKey and yKeys can not be trueful at the same time"
    assert y_key or y_keys, "yKey and yKeys can not be both falseful."
    assert head is None or tail is None, "head and tail can not be trueful at the same time"
    if not prefix:
        for id in metrics_files:
            assert isabs(id), f"metricFile need to be absolute path is prefix is {prefix}. It is {id} instead."

    ids = [join(prefix or "", id) for id in metrics_files]
    dfs = [read_dataframe(join(Args.logdir, _id[1:])) for _id in ids]

    y_keys = y_keys or [y_key]
    join_keys = [k for k in {x_key, *y_keys} if k is not None]

    dataframes = []
    for df in dfs:
        if df is None:
            df['index'] = df.index
            df.set_index('index')
            continue
        elif x_key is not None:
            df.set_index(x_key)
            if x_align is None:
                pass
            elif x_align == "start":  # todo: this needs to be part of the join
                df[x_key] -= df[x_key][0]
            elif x_align == "end":
                df[x_key] -= df[x_key][-1]
            else:
                df[x_key] -= x_align
        else:
            df = df[y_keys]
            df['index'] = df.index
            df.set_index('index')

        # todo: maybe apply tail and head *after* dropna??
        if tail is not None:
            df = df.tail(tail)
        if head is not None:
            df = df.head(head)
        inds = True
        if x_low is not None:
            inds &= df[x_key or "index"] >= x_low
            print("x_low >>>", inds)
        if x_high is not None:
            inds &= df[x_key or "index"] <= x_high
            print("x_high >>>", inds)
        if inds is not True:
            df = df.loc[inds]

        # todo: only dropna if we are not using ranges. <need to test>
        try:
            column = df[join_keys]
            if head is None and tail is None:
                dataframes.append(column.dropna())
            else:
                dataframes.append(column)
        except KeyError as e:
            raise KeyError(f"{join_keys} contain keys that is not in the dataframe. "
                           f"Keys available include {df.keys()}") from e

    if not dataframes:  # No dataframe, return `null`.
        return None

    all = pd.concat(dataframes)

    if x_key:
        all = all.set_index(x_key)

    all.rank(method='first')

    if k is not None:
        bins = pd.qcut(all.index, k, duplicates='drop')
        grouped = all.groupby(bins)
        # df
    else:
        grouped = all.groupby(level=0)

    # treat all numbers in bin as equal. For raw (not averaged, or averaged)
    grouped[y_keys].agg(['count', 'mean', 'min', 'max'])
    df = grouped[y_keys].describe(percentiles=[0.25, 0.75, 0.5, 0.05, 0.95]).reset_index()

    if k is not None:
        if 'index' not in df:
            df['index'] = df.index
        if x_edge == "right" or x_edge is None:
            df['__x'] = df['index'].apply(lambda r: r.right)
        elif x_edge == "left":
            df['__x'] = df['index'].apply(lambda r: r.left)
        elif x_edge == "mean":
            df['__x'] = df['index'].apply(lambda r: 0.5 * (r.left + r.right))
        # todo: use mode of each bin
        else:
            raise KeyError(f"x_edge {[x_edge]} should be OneOf['start', 'after', 'mid', 'mode']")
    else:
        df['__x'] = df.index

    return Series(metrics_files,
                  _df=df.sort_values(by="__x"),
                  metrics_files=metrics_files,
                  prefix=prefix,
                  x_key=x_key or "index",
                  y_key=y_key,
                  y_keys=y_keys,
                  label=label,
                  warning=warning)


SeriesArguments = dict(
    metrics_files=List(String, required=True),
    prefix=String(description="prefix to the metricFiles.", required=False),
    head=Int(description="the number of datapoints (for each metrics file) to take from the head-end"),
    tail=Int(description="the number of datapoints (for each metrics file) to take from the tail-end"),
    x_low=Float(description="the (inclusive) lower end of the x column"),
    x_high=Float(description="the (inclusive) higher end of the x column"),
    k=Int(required=False, description='the number of datapoints to return.'),
    x_align=String(description="a number (anchor point), 'start', 'end'"),
    x_key=String(),
    y_key=String(description="You can leave the xKey, but the yKey is required."),
    y_keys=List(String, description="Alternatively you can pass a list of keys to yKey*s*."),
    label=String(),
    warning=String(),
)

# if __name__ == "__main__":



from typing import List, Any


def assign(d1, d2):
    if not d2:
        return d1
    for k, v in d2.items():
        if isinstance(d1.get(k, None), dict):
            d1[k] = assign(d1[k], v)
        else:
            d1[k] = v
    return d1


if __name__ == "__main__":
    object1 = {"a": 1, "b": 2, "c": 3}
    object2 = assign({"c": 4, "d": 5}, object1)
    assert object2['c'] == 3
    assert object2['d'] == 5


def idot_keys(d, strict=True):
    for k, v in d.items():
        if isinstance(v, dict):
            if not strict:
                yield k
            for _ in idot_keys(v, strict):
                yield k + "." + _
        else:
            yield k


def dot_keys(d, strict=True):
    return [*idot_keys(d, strict)]


if __name__ == "__main__":
    object = {"a": 1, "b": 2, "c": 3, "child": {"a": 3, "grandchild": {'d': 8}}}
    assert dot_keys(object) == ['a', 'b', 'c', 'child.a', 'child.grandchild.d']
    assert dot_keys(object, strict=False) == ['a', 'b', 'c', 'child', 'child.a', 'child.grandchild',
                                              'child.grandchild.d']


def idot_flatten(d, ancestors: List[Any] = tuple()):
    """
    returns a flattened dictionary with the keys of the nexted dictionaries converted into dot-separated keys.

    :param d: map
    :return: flat map
    """
    for k, v in d.items():
        if isinstance(v, dict):
            for _k, _v in idot_flatten(v):
                yield k + "." + _k, _v
        else:
            yield k, v


def dot_flatten(d):
    # note: dictionaries are ordered by default in python 3.7.
    return dict(idot_flatten(d))


if __name__ == "__main__":
    object = {"a": 1, "b": 2, "c": 3, "child": {"a": 3, "grandchild": {'d': 8}}}
    assert list(dot_flatten(object).keys()) == ['a', 'b', 'c', 'child.a', 'child.grandchild.d']



ignored = ['.bzr', '$RECYCLE.BIN', '.DAV', '.DS_Store', '.git', '.hg', '.htaccess', '.htpasswd', '.Spotlight-V100',
           '.svn', '__MACOSX', 'ehthumbs.db', 'robots.txt', 'Thumbs.db', 'thumbs.tps']
datatypes = {'audio': 'm4a,mp3,oga,ogg,webma,wav',
             'archive': '7z,zip,rar,gz,tar',
             'image': 'gif,ico,jpe,jpeg,jpg,png,svg,webp',
             'pdf': 'pdf',
             'quicktime': '3g2,3gp,3gp2,3gpp,mov,qt',
             'source': 'atom,bat,bash,c,cmd,coffee,css,hml,js,json,java,less,markdown,md,php,pl,py,rb,rss,sass,scpt,swift,scss,sh,xml,yml,plist',
             'text': 'txt',
             'video': 'mp4,m4v,ogv,webm',
             'website': 'htm,html,mhtm,mhtml,xhtm,xhtml'}
icontypes = {'fa-music': 'm4a,mp3,oga,ogg,webma,wav',
             'fa-archive': '7z,zip,rar,gz,tar',
             'fa-picture-o': 'gif,ico,jpe,jpeg,jpg,png,svg,webp',
             'fa-file-text': 'pdf',
             'fa-film': '3g2,3gp,3gp2,3gpp,mov,qt',
             'fa-code': 'atom,plist,bat,bash,c,cmd,coffee,css,hml,js,json,java,less,markdown,md,php,pl,py,rb,rss,sass,scpt,swift,scss,sh,xml,yml',
             'fa-file-text-o': 'txt',
             'fa-film': 'mp4,m4v,ogv,webm',
             'fa-globe': 'htm,html,mhtm,mhtml,xhtm,xhtml'}



import pytest
from graphene.test import Client
from graphql_relay import to_global_id
from ml_dash.schema import schema
from dash_server_specs import show, shows


@pytest.fixture(scope='session')
def log_dir(request):
    return request.config.getoption('--log-dir')


def test_delete_text_file(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        mutation AppMutation ($id: ID!) {
            deleteFile (input: { 
                            id: $id, 
                            clientMutationId: "10", 
             }) { ok }
        }
    """
    path = "/episodeyang/cpc-belief/README.md"
    r = client.execute(query, variables=dict(id=to_global_id("File", path)))

    if 'errors' in r:
        raise RuntimeError("\n" + shows(r['errors']))
    else:
        print(">>")
        show(r['data'])


def test_mutate_text_file(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        mutation AppMutation ($id: ID!) {
            updateText (input: { 
                            id: $id, 
                            text: "new text!!\\n1\\n2\\n3\\n4\\n5\\n6",
                            clientMutationId: "10", 
             }) { 
                file { id name text (stop:5) }
            }
        }
    """
    path = "/episodeyang/cpc-belief/README.md"
    r = client.execute(query, variables=dict(id=to_global_id("File", path)))

    if 'errors' in r:
        raise RuntimeError("\n" + shows(r['errors']))
    else:
        print(">>")
        show(r['data'])


def test_mutate_json_file(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        mutation AppMutation ($id: ID!) {
            updateJson (input: { 
                            id: $id, 
                            data: {text: "hey", key: 10},
                            clientMutationId: "10", 
             }) { 
                file { id name json text yaml }
            }
        }
    """
    path = "/episodeyang/cpc-belief/README.md"
    r = client.execute(query, variables=dict(id=to_global_id("File", path)))

    if 'errors' in r:
        raise RuntimeError("\n" + shows(r['errors']))
    else:
        print(">>")
        show(r['data'])


def test_mutate_yaml_file(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        mutation AppMutation ($id: ID!) {
            updateYaml (input: { 
                            id: $id, 
                            data: {text: "hey", key: 10, charts: [0, 1]},
                            clientMutationId: "10", 
             }) { 
                file { id name text yaml }
            }
        }
    """
    path = "/episodeyang/cpc-belief/README.md"
    r = client.execute(query, variables=dict(id=to_global_id("File", path)))

    if 'errors' in r:
        raise RuntimeError("\n" + shows(r['errors']))
    else:
        print(">>")
        show(r['data'])


def test_glob_files(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        query AppQuery ($cwd: String!, $query: String) {
            glob ( cwd: $cwd, query: $query) { id name path }
        }
    """
    path = "/episodeyang/cpc-belief/mdp/experiment_00/"
    r = client.execute(query, variables=dict(cwd=path, query="figures/*.png"))
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        print(">>")
        show(r['data'])


def test_experiment(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        query AppQuery ($id: ID!) {
            experiment ( id:  $id ) { 
                id
                name 
                parameters { id name }
                files (first:10) {
                    edges {
                        node {
                            id name path
                        }
                    }
                }
                directories (first:10) {
                    edges {
                        node {
                            id name path
                        }
                    }
                }
            }
        }
    """
    path = "/episodeyang/cpc-belief/mdp/experiment_01"
    r = client.execute(query, variables=dict(id=to_global_id("Experiment", path)))
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        print(">>")
        show(r['data'])


def test_directory(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        query AppQuery ($id: ID!) {
            directory ( id:  $id ) { 
                id
                name 
                path
                readme {
                    id name path 
                    text(stop:11)
                }
                dashConfigs(first:10) {
                    edges {
                        node {
                            id name 
                            path 
                            yaml
                            text(stop:11)
                        }
                    }
                }
                
                charts(first:10) {
                    edges {
                        node {
                            id name 
                            dir 
                            path 
                            yaml
                            text(stop:11)
                        }
                    }
                }
                
                directories (first:10) {
                    edges {
                        node {
                            id name path
                            directories (first:10) {
                                edges {
                                    node {
                                        id name
                                    }
                                }
                            }
                        }
                    }
                }
                experiments (first:10) {
                    edges { node { 
                        id name path
                        parameters {keys flat}
                        files (first:10) { edges { node { id, name} } }
                    } }
                }
            }
        }
    """
    path = "/episodeyang/cpc-belief/mdp"
    r = client.execute(query, variables=dict(id=to_global_id("Directory", path)))
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        print(">>")
        show(r['data'])


# todo: add chunked loading for the text field. Necessary for long log files.
def test_reac_text_file(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        query AppQuery ($id: ID!) {
            file ( id:  $id) { 
                id name text
            }
        }
    """
    path = "/episodeyang/cpc-belief/README.md"
    r = client.execute(query, variables=dict(id=to_global_id("File", path)))
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        print(">>")
        show(r['data'])


def test_series_2(log_dir):
    query = """
    query LineChartsQuery(
      $prefix: String
      $xKey: String
      $yKey: String
      $yKeys: [String]
      $metricsFiles: [String]!
    ) {
      series(metricsFiles: $metricsFiles, prefix: $prefix, k: 10, xKey: $xKey, yKey: $yKey, yKeys: $yKeys) {
        id
        xKey
        yKey
        xData
        yMean
        yCount
      }
    }
    """
    variables = {"prefix": None, "xKey": "epoch", "yKey": "slow_sine", "yKeys": None,
                 "metricsFiles": ["/episodeyang/cpc-belief/mdp/experiment_01/metrics.pkl"]}

    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    r = client.execute(query, variables=variables)
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        print(">>")
        show(r['data'])

    assert r['data']['series']['xData'].__len__() == 10


def test_series_x_limit(log_dir):
    query = """
    query LineChartsQuery(
      $prefix: String
      $xKey: String
      $yKey: String
      $yKeys: [String]
      $metricsFiles: [String]!
    ) {
      series(
        metricsFiles: $metricsFiles, 
        prefix: $prefix, 
        k: 10, 
        xLow: 100,
        xKey: $xKey, 
        yKey: $yKey, 
        yKeys: $yKeys
        ) {
            id
            xKey
            yKey
            xData
            yMean
            yCount
          }
    }
    """

    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)

    variables = {"prefix": None, "xKey": "epoch", "yKey": "slow_sine", "yKeys": None,
                 "metricsFiles": ["/episodeyang/cpc-belief/mdp/experiment_01/metrics.pkl"]}
    r = client.execute(query, variables=variables)
    print(r['data']['series']['xData'])
    assert r['data']['series']['xData'].__len__() == 10

    variables = {"prefix": None, "yKey": "slow_sine", "yKeys": None,
                 "metricsFiles": ["/episodeyang/cpc-belief/mdp/experiment_01/metrics.pkl"]}
    r = client.execute(query, variables=variables)
    print(r['data']['series']['xData'])
    assert r['data']['series']['xData'].__len__() == 10


def test_series_last(log_dir):
    query = """
    query LastMetricQuery(
      $yKey: String
      $last: Int
      $metricsFiles: [String]!
    ) {
      series(metricsFiles: $metricsFiles, k: 1, yKey: $yKey, tail: $last) {
        id
        yKey
        yMean
        yCount
      }
    }
    """
    variables = {"yKey": "slow_sine",
                 "last": 100,
                 "metricsFiles": ["/episodeyang/cpc-belief/mdp/experiment_01/metrics.pkl"]}

    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    r = client.execute(query, variables=variables)
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        print(">>")
        show(r['data'])
        assert not not r['data']['series']['yMean'], "the yMean should NOT be empty"
        assert not not r['data']['series']['yCount'] == [100.0]


# can we do the average first?
def test_series_group(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        query AppQuery {
            series (
                k:30
                tail: 100
                xLow: 25
                prefix:"/episodeyang/cpc-belief/mdp"
                metricsFiles:["experiment_00/metrics.pkl", "experiment_01/metrics.pkl", "experiment_02/metrics.pkl"]
                # xKey: "__timestamp"
                xKey: "epoch"
                # yKey: "sine"
                yKeys: ["sine", "slow_sine"]
                # xAlign: "start"
            ) { 
                id
                xKey
                yKey
                yKeys
                xData
                yMean
                yMedian
                y25pc
                y75pc
                y05pc
                y95pc
                yMedian
                yCount
            }
        }
    """
    r = client.execute(query, variables=dict(username="episodeyang"))
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        print(">>")
        show(r['data'])


def test_metric(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        query AppQuery {
            metrics(id:"TWV0cmljczovZXBpc29kZXlhbmcvY3BjLWJlbGllZi9tZHAvZXhwZXJpbWVudF8wMC9tZXRyaWNzLnBrbA==" ) { 
                id
                keys
                path
                value (keys: ["__timestamp", "sine"])
            }
        }
    """
    r = client.execute(query, variables=dict(username="episodeyang"))
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        print(">>")
        show(r['data'])


def test_schema(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        query AppQuery ($username:String) {
            user (username: $username) {
                username 
                name 
                projects(first: 1) {
                  edges {
                    node {
                        id
                        name
                        experiments(first:10) { edges { node { 
                            id
                            name
                            parameters {value keys flat raw} 
                            metrics {
                                id
                                keys 
                                value (keys: ["__timestamp", "sine"]) 
                            }
                        } } }
                        directories(first:10){
                            edges {
                                node {
                                    id
                                    name
                                    files(first:10){
                                        edges {
                                            node { name }
                                        }
                                    }
                                    directories(first:10){
                                        edges {
                                            node { name }
                                        }
                                    }
                                }
                            }
                        }
                    }
                  }
                }
            }
        }
    """
    r = client.execute(query, variables=dict(username="episodeyang"))
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        _ = r['data']['user']
        show(_)


def test_node(log_dir):
    from ml_dash.config import Args
    Args.logdir = log_dir
    client = Client(schema)
    query = """
        query AppQuery ($id:ID!) {
            node (id: $id) {
                id
                ... on File {
                    name
                    text
                }
            }
        }
    """
    path = "/episodeyang/cpc-belief/README.md"
    r = client.execute(query, variables=dict(id=to_global_id("File", path)))
    if 'errors' in r:
        raise RuntimeError(r['errors'])
    else:
        print(">>")
        show(r['data'])



import os
from sanic import Sanic
from sanic.exceptions import FileNotFound
from sanic.response import file
from params_proto import cli_parse, Proto

# gets current directory
BASE = os.path.realpath(__file__)
build_path = os.path.join(os.path.dirname(BASE), "client-dist")

app = Sanic("ml_dash.app")
# serve js file for webpack
app.static('/', build_path)


@app.route('/')
@app.exception(FileNotFound)
async def index(request, exception=None):
    print('hey ====', [exception])
    return await file(build_path + '/index.html')


@cli_parse
class AppServerArgs:
    """
    Configuration Arguments for the Sanic App that serves
    the static web-application.

    [Usage]

    To launch the web-app client, do

    python -m ml_dash.app port=3001 host=0.0.0.0 workers=4 debug=True
    """
    host = Proto("", help="use 0.0.0.0 if you want external clients to be able to access this.")
    port = Proto(3001, help="the port")
    workers = Proto(1, help="the number of worker processes")
    debug = False
    access_log = True


if __name__ == '__main__':
    import socket
    from termcolor import cprint, colored as c

    if AppServerArgs.host:
        from requests import get

        host_ip = get('https://api.ipify.org').text
        del get
        ip_string = f"""
      Local:       {c(f'http://localhost:{AppServerArgs.port}/', 'green')}
      External Ip: {c(f'http://{host_ip}:{AppServerArgs.port}/', 'green')}"""
    else:
        ip_string = f"""
      Local: {c(f'http://localhost:{AppServerArgs.port}/', 'green')}"""
    # try:
    #     hostname = socket.gethostname()
    #     host_ip = socket.gethostbyname(hostname)
    # except Exception as e:
    #     host_ip = "<not available>"

    print(f"""
    You can now view {c('ml-dash client', 'white')} in the browser.
{ip_string}

    To update to the newer version, do 
    {c('~>', 'blue')} {c('pip install --upgrade ml-dash', 'red')}
        
    """)
    app.run(**vars(AppServerArgs))



import os
from os.path import split, isabs, realpath, join, basename, dirname
from graphene import ObjectType, relay, String, Int, Mutation, ID, Field, Node, Boolean
from graphene.types.generic import GenericScalar
from graphql_relay import from_global_id
from ml_dash.schema.files.file_helpers import find_files

from . import parameters, metrics


class File(ObjectType):
    class Meta:
        interfaces = relay.Node,

    dir = String(description="the parent directory")
    name = String(description='name of the file')
    stem = String(description="stem of the file name")
    path = String(description='path to the file')

    def resolve_dir(self, info):
        return os.path.dirname(self.id)

    def resolve_stem(self, info, ):
        return self.name.split('.')[0]

    def resolve_path(self, info):
        return self.id

    text = String(description='text content of the file',
                  start=Int(required=False, default_value=0),
                  stop=Int(required=False, default_value=None))

    def resolve_text(self, info, start=0, stop=None):
        from ml_dash.config import Args
        try:
            with open(join(Args.logdir, self.id[1:]), "r") as f:
                lines = list(f)[start: stop]
                return "".join(lines)
        except FileNotFoundError:
            return None

    json = GenericScalar(description="the json content of the file")

    def resolve_json(self, info):
        import json
        try:
            from ml_dash.config import Args
            with open(join(Args.logdir, self.id[1:]), "r") as f:
                return json.load(f)
        except FileNotFoundError:
            return None

    yaml = GenericScalar(description="the content of the file using yaml")

    def resolve_yaml(self, info):
        import ruamel.yaml
        if ruamel.yaml.version_info < (0, 15):
            yaml = ruamel.yaml
            load_fn = yaml.safe_load
        else:
            from ruamel.yaml import YAML
            yaml = YAML()
            yaml.explict_start = True
            load_fn = yaml.load

        from ml_dash.config import Args
        try:
            with open(join(Args.logdir, self.id[1:]), "r") as f:
                return load_fn('\n'.join(f))
        except FileNotFoundError:
            return None

    @classmethod
    def get_node(cls, info, id):
        return get_file(id)


class FileConnection(relay.Connection):
    class Meta:
        node = File


def get_file(id):
    return File(id=id, name=basename(id), path=id)


def find_files_by_query(cwd, query="**/*.*", **kwargs):
    from ml_dash.config import Args
    assert isabs(cwd), "the current work directory need to be an absolute path."
    _cwd = realpath(join(Args.logdir, cwd[1:])).rstrip('/')
    parameter_files = find_files(_cwd, query, **kwargs)
    return [
        # note: not sure about the name.
        File(id=join(cwd.rstrip('/'), p['path']),
             name=basename(p['path']),
             path=join(cwd.rstrip('/'), p['path']))
        for p in parameter_files
    ]


def save_text_to_file(path, text):
    from ml_dash.config import Args
    assert isabs(path), "the path has to be absolute path."
    _path = join(Args.logdir, path[1:])
    with open(_path, "w") as f:
        f.write(text)
    return get_file(path)


def save_yaml_to_file(path, data):
    from ml_dash.config import Args
    assert isabs(path), "the path has to be absolute path."
    _path = join(Args.logdir, path[1:])
    # note: assume all text format
    with open(_path, "w+") as f:
        import yaml
        _ = yaml.dump(data, f)
    return get_file(path)


def save_json_to_file(path, data):
    from ml_dash.config import Args
    assert isabs(path), "the path has to be absolute path."
    _path = join(Args.logdir, path[1:])
    # note: assume all text format
    with open(_path, "w+") as f:
        import json
        _ = json.dumps(data, sort_keys=True, indent=2)
        f.write(_)
    return get_file(path)


def remove_file(path):
    """remove does not work with directories"""
    from ml_dash.config import Args
    assert isabs(path), "the path has to be absolute path."
    _path = join(Args.logdir, path[1:])
    os.remove(_path)


def remove_directory(path):
    """rmtree does not work with files"""
    import shutil
    from ml_dash.config import Args
    assert isabs(path), "the path has to be absolute path."
    _path = join(Args.logdir, path[1:])
    shutil.rmtree(_path)


class MutateTextFile(relay.ClientIDMutation):
    class Input:
        id = ID()
        text = String(required=True)

    file = Field(File)

    @classmethod
    def mutate_and_get_payload(cls, root, info, id, text, client_mutation_id):
        _type, path = from_global_id(id)
        return MutateTextFile(file=save_text_to_file(path, text))


class MutateYamlFile(relay.ClientIDMutation):
    """
    serializes the data to a yaml file format
    """

    class Input:
        id = ID()
        data = GenericScalar()

    file = Field(File)

    @classmethod
    def mutate_and_get_payload(self, root, info, id, data, client_mutation_id):
        _type, path = from_global_id(id)
        return MutateYamlFile(file=save_yaml_to_file(path, data))


class MutateJSONFile(relay.ClientIDMutation):
    """
    serializes the data to a json file format
    """

    class Input:
        id = ID()
        data = GenericScalar()

    file = Field(File)

    @classmethod
    def mutate_and_get_payload(self, root, info, id, data, client_mutation_id):
        _type, path = from_global_id(id)
        return MutateJSONFile(file=save_json_to_file(path, data))


class DeleteFile(relay.ClientIDMutation):
    class Input:
        id = ID()

    ok = Boolean()
    id = ID()

    @classmethod
    def mutate_and_get_payload(cls, root, info, id, client_mutation_id):
        _type, path = from_global_id(id)
        try:
            remove_file(path)
            return DeleteFile(ok=True, id=id)
        except FileNotFoundError:
            return DeleteFile(ok=False)


class DeleteDirectory(relay.ClientIDMutation):
    class Input:
        id = ID()

    ok = Boolean()
    id = ID()

    @classmethod
    def mutate_and_get_payload(cls, root, info, id, client_mutation_id):
        _type, path = from_global_id(id)
        try:
            remove_directory(path)
            return DeleteDirectory(ok=True, id=id)
        except FileNotFoundError:
            return DeleteDirectory(ok=False)



from os import listdir
from os.path import isfile, join, split
from graphene import ObjectType, relay, String, Field
from ml_dash import schema


class Directory(ObjectType):
    class Meta:
        interfaces = relay.Node,

    name = String(description='name of the directory')
    path = String(description='absolute path of the directory')

    readme = Field(lambda: schema.files.File)

    def resolve_readme(self, info, *args, **kwargs):
        # note: keep it simple, just use README for now.
        readmes = schema.files.find_files_by_query(cwd=self.id, query="README.md")
        return readmes[0] if readmes else None

    # deprecate this
    dash_configs = relay.ConnectionField(lambda: schema.files.FileConnection)

    def resolve_dash_configs(self, info, *args, **kwargs):
        return schema.files.find_files_by_query(cwd=self.id, query="*.dashcfg")

    charts = relay.ConnectionField(lambda: schema.files.FileConnection)

    def resolve_charts(self, info, *args, **kwargs):
        return schema.files.find_files_by_query(cwd=self.id, query="**/*.chart.yml")

    experiments = relay.ConnectionField(lambda: schema.experiments.ExperimentConnection)

    def resolve_experiments(self, info, first=None, **kwargs):
        if first is not None:
            return schema.experiments.find_experiments(cwd=self.id, stop=first)
        return schema.experiments.find_experiments(cwd=self.id)

    directories = relay.ConnectionField(lambda: schema.directories.DirectoryConnection)

    def resolve_directories(self, info, **kwargs):
        from ml_dash.config import Args
        root_dir = join(Args.logdir, self.id[1:])
        return [get_directory(join(self.id, _))
                for _ in listdir(root_dir) if not isfile(join(root_dir, _))]

    files = relay.ConnectionField(lambda: schema.files.FileConnection)

    def resolve_files(self, info, **kwargs):
        from ml_dash.config import Args
        root_dir = join(Args.logdir, self.id[1:])
        return [schema.files.File(id=join(self.id, _), name=_)
                for _ in listdir(root_dir) if isfile(join(root_dir, _))]

    @classmethod
    def get_node(cls, info, id):
        return get_directory(id)


class DirectoryConnection(relay.Connection):
    class Meta:
        node = Directory


def get_directory(id):
    _id = id.rstrip('/')
    return Directory(id=_id, name=split(_id[1:])[-1], path=_id)



from ml_dash.schema import schema
from sanic_graphql import GraphQLView

from sanic import Sanic, views
from sanic_cors import CORS

# to support HTTPS.
views.HTTP_METHODS += ('FETCH', 'OPTIONS')

app = Sanic("ml_dash.server")
# CORS(app)
CORS(app, resources={r"/*": {"origins": "*"}}, automatic_options=True)

# NOTE: disable for laziness. Should enable it in the future.
# @app.listener('before_server_start')
# def init_graphql(app, loop):
#     app.add_route(GraphQLView.as_view(schema=schema, executor=AsyncioExecutor(loop=loop)), '/graphql')

# new graphQL endpoints
app.add_route(GraphQLView.as_view(schema=schema, graphiql=True), '/graphql',
              methods=['GET', 'POST', 'FETCH', 'OPTIONS'])
app.add_route(GraphQLView.as_view(schema=schema, batch=True), '/graphql/batch',
              methods=['GET', 'POST', 'FETCH', 'OPTIONS'])


@app.listener('before_server_start')
def setup_static(app, loop):
    from ml_dash import config
    from os.path import expanduser
    app.static('/files', expanduser(config.Args.logdir),
               use_modified_since=True, use_content_range=True, stream_large_files=True)


# note: currently disabled, file events API.
# from .file_events import file_events, setup_watch_queue
# app.add_route(file_events, '/file-events', methods=['GET', 'OPTIONS'])
# app.add_route(file_events, '/file-events/<file_path:path>', methods=['GET', 'OPTIONS'])
# app.listener('before_server_start')(setup_watch_queue)


def run(logdir=None, **kwargs):
    from ml_dash import config
    from termcolor import cprint

    if logdir:
        config.Args.logdir = logdir

    cprint("launched server with config:", "green")
    cprint("Args:", 'yellow')
    print(vars(config.Args))
    cprint("Sanic Server Args:", 'yellow')
    print(vars(config.ServerArgs))
    cprint("SSL Args:", 'yellow')
    print(vars(config.SSLArgs))

    config.ServerArgs.update(**kwargs)
    if config.SSLArgs:
        app.run(**vars(config.ServerArgs), ssl=vars(config.SSLArgs))
    else:
        app.run(**vars(config.ServerArgs))


if __name__ == "__main__":
    # see: https://sanic.readthedocs.io/en/latest/sanic/deploying.html
    # call this as `python -m ml_logger.main`
    run()



from functools import reduce
from os.path import split, join as pJoin, basename, realpath
from graphene import ObjectType, relay, String, List
from graphene.types.generic import GenericScalar
from ml_dash.config import Args
from ml_dash.schema.files.file_helpers import find_files, read_pickle_for_json
from ml_dash.schema.helpers import assign, dot_keys, dot_flatten


class Parameters(ObjectType):
    class Meta:
        interfaces = relay.Node,

    name = String(description="The true path to the parameter file. Internal use only")
    path = String(description="The true path to the parameter file. Internal use only")
    keys = List(String, description="list of parameter keys")
    value = GenericScalar(description="the json value for the parameters")
    raw = GenericScalar(description="the raw data object for the parameters")
    flat = GenericScalar(description="the raw data object for the parameters")

    def resolve_name(self, info):
        return basename(self.id)

    def resolve_path(self, info):
        return self.id

    def resolve_keys(self, info):
        value = reduce(assign, read_pickle_for_json(pJoin(Args.logdir, self.id[1:])) or [{}])
        return dot_keys(value)

    def resolve_value(self, info, **kwargs):
        return reduce(assign, read_pickle_for_json(pJoin(Args.logdir, self.id[1:])) or [{}])

    def resolve_raw(self, info, **kwargs):
        return read_pickle_for_json(pJoin(Args.logdir, self.id[1:]))

    def resolve_flat(self, info, **kwargs):
        # note: this always gives truncated some-folder/arameter.pkl path.
        value = reduce(assign, read_pickle_for_json(pJoin(Args.logdir, self.id[1:])) or [{}])
        return dot_flatten(value)

    # description = String(description='string serialized data')
    # experiments = List(lambda: schema.Experiments)

    @classmethod
    def get_node(cls, info, id):
        return get_parameters(id)


class ParameterConnection(relay.Connection):
    class Meta:
        node = Parameters


def get_parameters(id):
    return Parameters(id=id)


def find_parameters(cwd, **kwargs):
    from ml_dash.config import Args
    _cwd = realpath(pJoin(Args.logdir, cwd[1:]))
    parameter_files = find_files(_cwd, "parameters.pkl", **kwargs)
    for p in parameter_files:
        yield Parameters(id=pJoin(cwd, p['path']))



def bind(fn):
    """
    Binds the function to the class.

    :param fn:
    :return: bound_fn
    """
    return lambda _, *args, **kwargs: fn(*args, **kwargs)


def bind_args(fn):
    """
    Binds args after info.

    :param fn:
    :return: bound_fn
    """
    return lambda _, info, *args, **kwargs: fn(*args, **kwargs)




from graphene import relay, ObjectType, Float, Schema, List, String, Field, Int
from ml_dash.schema.files.series import Series, get_series, SeriesArguments
from ml_dash.schema.files.metrics import Metrics, get_metrics
from ml_dash.schema.schema_helpers import bind, bind_args
from ml_dash.schema.users import User, get_users, get_user
from ml_dash.schema.projects import Project
from ml_dash.schema.directories import Directory, get_directory
from ml_dash.schema.files import File, FileConnection, MutateTextFile, MutateJSONFile, MutateYamlFile, \
    DeleteFile, DeleteDirectory, find_files_by_query
# MutateJSONFile, MutateYamlFile
from ml_dash.schema.experiments import Experiment


class EditText(relay.ClientIDMutation):
    class Input:
        text = String(required=True, description='updated content for the text file')

    text = String(description="the updated content for the text file")

    @classmethod
    def mutate_and_get_payload(cls, root, info, text, ):
        return dict(text=text)


class Query(ObjectType):
    node = relay.Node.Field()
    # context?
    # todo: files
    # todo: series

    users = Field(List(User), resolver=bind_args(get_users))
    user = Field(User, username=String(), resolver=bind_args(get_user))
    series = Field(Series, resolver=bind_args(get_series), **SeriesArguments)

    project = relay.Node.Field(Project)
    experiment = relay.Node.Field(Experiment)
    metrics = relay.Node.Field(Metrics)
    directory = relay.Node.Field(Directory)
    file = relay.Node.Field(File)

    glob = Field(List(File), cwd=String(required=True), query=String(), start=Int(), stop=Int(),
                 resolver=bind_args(find_files_by_query))


class Mutation(ObjectType):
    # todo: create_file
    # done: edit_file
    # done: remove_file
    # todo: move_file
    # todo: copy_file

    # do we need to have separate deleteDirectory? (look up relay client-side macros)

    delete_file = DeleteFile.Field()
    delete_directory = DeleteDirectory.Field()
    # update_text = EditText.Field()
    update_text = MutateTextFile.Field()
    update_json = MutateJSONFile.Field()
    update_yaml = MutateYamlFile.Field()


schema = Schema(query=Query, mutation=Mutation)



from os.path import split, realpath, join, splitext, basename
from graphene import relay, ObjectType, String, List, JSONString, Int
from graphene.types.generic import GenericScalar
from ml_dash.config import Args
from ml_dash.schema.files.file_helpers import find_files, read_records, read_dataframe


class Metrics(ObjectType):
    class Meta:
        interfaces = relay.Node,

    path = String(description="path to the file")
    name = String(description="path to the file")
    stem = String(description="stem of the file name")

    def resolve_path(self, info):
        return self.id

    def resolve_stem(self, info, ):
        return self.name.split('.')[0]

    keys = List(String, description="list of keys for the metrics")

    # value = List(GenericScalar, description="the raw value")
    value = GenericScalar(description="The value of the metrics file",
                          keys=List(String),
                          k=Int(required=False),
                          last=Int(required=False),
                          window=Int(required=False))

    def resolve_keys(self, info):
        df = read_dataframe(join(Args.logdir, self.id[1:]))
        keys = df.keys()
        return list(keys)

    # todo: add more complex queries.
    def resolve_value(self, info, keys=None, k=None, last=None, window=None):
        path = join(Args.logdir, self.id[1:])
        realpath(path)
        _ = read_dataframe(path)
        if keys:
            df = _[keys].dropna()
            return {k: df[k].values.tolist() for k in keys}
        else:
            df = _.dropna()
            return {k: v.values.tolist() for k, v in df.items()}

    @classmethod
    def get_node(cls, info, id):
        return Metrics(id, )


class MetricsConnection(relay.Connection):
    class Meta:
        node = Metrics


def get_metrics(id):
    return Metrics(id=id, name=basename(id), path=id)


def find_metrics(cwd, **kwargs):
    from ml_dash.config import Args
    _cwd = realpath(join(Args.logdir, cwd[1:]))
    parameter_files = find_files(_cwd, "**/metrics.pkl", **kwargs)
    for p in parameter_files:
        yield Metrics(id=join(cwd, p['path']), name="metrics.pkl")



from os.path import expanduser

import numpy as np

from ml_logger import logger

DEBUG_DIR = expanduser('~/ml-logger-debug')

if __name__ == "__main__":
    from scipy.misc import face

    fn = lambda x: np.random.rand() + (1 + 0.001 * x) * np.sin(x * 0.1 / np.pi)
    fn_1 = lambda x: np.random.rand() + (1 + 0.001 * x) * np.sin(x * 0.04 / np.pi)

    for username in ["episodeyang", "amyzhang"]:
        for project in ['cpc-belief', 'playground']:
            for i in range(2):
                prefix = f"{username}/{project}/{'mdp/' if i < 5 else '/'}experiment_{i:02d}"
                logger.remove(prefix)

                logger.configure(log_directory=DEBUG_DIR, prefix=prefix)

                logger.log_params(Args=dict(lr=10 ** (-2 - i),
                                            weight_decay=0.001,
                                            gradient_clip=0.9,
                                            env_id="GoalMassDiscreteIdLess-v0",
                                            seed=int(i * 100)))
                for ep in range(50 + 1):
                    logger.log_metrics(epoch=ep, sine=fn(ep), slow_sine=fn_1(ep))
                    logger.flush()
                    if ep % 10 == 0:
                        logger.save_image(face('gray'), f"figures/gray_{ep:04d}.png")
                        logger.save_image(face('rgb'), f"figures/rgb_{ep:04d}.png")

                logger.save_image(face('gray'), "figures/face_gray.png")
                logger.save_image(face('rgb'), "figures/face_rgb.png")

            with logger.PrefixContext(f"runs/{username}/{project}"):
                logger.log_line("# Root Files\n", file="RAEDME.md")

    # import numpy as np
    # import matplotlib.pyplot as plt
    #
    # xs = np.arange(500)
    # ys = (1 + 0.001 * xs) * np.sin(xs * 0.1 / np.pi)
    # ys += np.random.rand(*ys.shape) * 1
    # plt.plot(xs, ys)
    # plt.show()
    #
    # import numpy as np
    # import matplotlib.pyplot as plt
    #
    # xs = np.arange(500)
    # ys = (1 + 0.001 * xs) * np.sin(xs * 0.02 / np.pi)
    # ys += np.random.rand(*ys.shape) * 1
    # plt.plot(xs, ys)
    # plt.show()



# from . import config
import asyncio
from hachiko.hachiko import AIOWatchdog


class Handler:
    def dispatch(self, *args, **kwargs):
        print(args, kwargs)

@asyncio.coroutine
def watch_fs(path):
    watch = AIOWatchdog(path, event_handler=Handler())
    watch.start()
    while True:
        yield from asyncio.sleep(10)
    watch.stop()



if __name__ == "__main__":
    # asyncio.get_event_loop().run_until_complete(watch_fs("/Users/ge/machine_learning/berkeley-playground/ins-runs"))
    # asyncio.get_event_loop().run_until_complete(watch_fs("."))
    path = "."
    watch = AIOWatchdog(path, event_handler=Handler())
    watch.start()
    import time
    print('watch is setup')
    while True:
        time.sleep(10)




import os

from params_proto import cli_parse, Proto


@cli_parse
class Args:
    """
    ML-Dash
    -------

    This module contains `ml_dash.server`, the visualization backend, and `ml_dash.app`, a
    static server hosting the web application.

    Usage
    -----

        python -m ml_dash.server --port 8090 --host 0.0.0.0 --workers 10

    """
    logdir = Proto(os.path.realpath("."), help="the root directory for all of the logs")


@cli_parse
class ServerArgs:
    host = Proto("", help="use 0.0.0.0 if you want external clients to be able to access this.")
    port = Proto(8081, help="the port")
    workers = Proto(1, help="the number of worker processes")
    debug = False

@cli_parse
class SSLArgs:
    cert = Proto(None, dtype=str, help="the path to the SSL certificate")
    key = Proto(None, dtype=str, help="the path to the SSL key")



import pathlib
from glob import iglob
from os import stat
from os.path import basename, join, realpath, dirname

from ml_dash.file_handlers import cwdContext


def file_stat(file_path, no_stat=True):
    """
    getting the stats of the file.

    no_stat turns the stat call off.

    :param file_path:
    :param no_stat:
    :return:
    """
    # note: this when looped over is very slow. Fine for a small list of files though.
    if no_stat:
        return dict(
            name=basename(file_path),
            path=file_path,
            dir=dirname(file_path),
        )

    stat_res = stat(file_path)
    sz = stat_res.st_size
    return dict(
        name=basename(file_path),
        path=file_path,
        dir=dirname(file_path),
        time_modified=stat_res.st_mtime,
        time_created=stat_res.st_ctime,
        # type=ft,
        size=sz,
    )


def fast_glob(query, wd, skip_children=False):
    """
    ignore subtree when file is found under a certain directory.
    :param skip_childre:
    :return:
    """
    raise NotImplementedError()


def find_files(cwd, query, start=None, stop=None, no_stat=True, show_progress=False):
    """
    find files by iGlob.

    :param cwd: the context folder for the glob, excluded from returned path list.
    :param query: glob query
    :param start: starting index for iGlob.
    :param stop: ending index for iGlob
    :param no_stat: boolean flag to turn off the file_stat call.
    :return:
    """
    from itertools import islice

    # https://stackoverflow.com/a/58126417/1560241
    if query.endswith('**'):
        query += "/*"

    with cwdContext(cwd):
        _ = islice(pathlib.Path(".").glob(query), start, stop)
        if show_progress:
            from tqdm import tqdm
            _ = tqdm(_, desc="@find_files")
        for i, file in enumerate(_):
            print(str(file))
            yield file_stat(str(file), no_stat=no_stat)


def read_dataframe(path, k=None):
    from ml_logger.helpers import load_pickle_as_dataframe
    try:
        return load_pickle_as_dataframe(path, k)
    except FileNotFoundError:
        return None


def read_records(path, k=200):
    from ml_logger.helpers import load_pickle_as_dataframe
    df = load_pickle_as_dataframe(path, k)
    return df.to_json(orient="records")


def read_log(path, k=200):
    from ml_logger.helpers import load_pickle_as_dataframe
    df = load_pickle_as_dataframe(path, k)
    return df.to_json(orient="records")


def read_pikle(path):
    from ml_logger.helpers import load_from_pickle
    data = [_ for _ in load_from_pickle(path)]
    return data


def read_pickle_for_json(path):
    """convert non JSON serializable types to string"""
    from ml_logger.helpers import load_from_pickle, regularize_for_json
    data = [regularize_for_json(_) for _ in load_from_pickle(path)]
    return data


def read_text(path, start, stop):
    from itertools import islice
    with open(path, 'r') as f:
        text = ''.join([l for l in islice(f, start, stop)])
    return text


def read_binary():
    raise NotImplementedError()
    # todo: check the file handling here. Does this use correct
    #  mimeType for text files?
    # res = await response.file(path)
    # if as_attachment:
    #     res.headers['Content-Disposition'] = 'attachment'



from os import listdir
from os.path import isfile, join, split

from graphene import ObjectType, relay, String, List
from ml_dash import schema


class Project(ObjectType):
    class Meta:
        interfaces = relay.Node,

    name = String(description='name of the project')

    # description = String(description='string serialized data')
    # experiments = List(lambda: schema.Experiments)

    experiments = relay.ConnectionField(lambda: schema.experiments.ExperimentConnection)

    def resolve_experiments(self, info, before=None, after=None, first=None, last=None):
        # todo: add support for before after and last
        if first is not None:
            return schema.experiments.find_experiments(cwd=self.id, stop=first)
        return schema.experiments.find_experiments(cwd=self.id)

    directories = relay.ConnectionField(lambda: schema.directories.DirectoryConnection)

    def resolve_directories(self, info, before=None, after=None, first=None, last=None):
        from ml_dash.config import Args
        root_dir = join(Args.logdir, self.id[1:])
        return [schema.Directory(id=join(self.id, _), name=_)
                for _ in listdir(root_dir) if not isfile(join(root_dir, _))]

    files = relay.ConnectionField(lambda: schema.files.FileConnection)

    def resolve_files(self, info, before=None, after=None, first=None, last=None):
        from ml_dash.config import Args
        root_dir = join(Args.logdir, self.id[1:])
        return [schema.Directory(id=join(self.id, _), name=_)
                for _ in listdir(root_dir) if isfile(join(root_dir, _))]

    @classmethod
    def get_node(cls, info, id):
        return get_project(id)


class ProjectConnection(relay.Connection):
    class Meta:
        node = Project


def get_projects(username):
    import os
    from ml_dash.config import Args
    user_root = join(Args.logdir, username)
    return [Project(name=_, id=join('/', username, _))
            for _ in os.listdir(user_root) if not isfile(_)]


def get_project(id):
    from ml_dash.config import Args
    path = join(Args.logdir, id[1:])
    return Project(id=id, name=split(id[1:])[1], _path=path)



from os import listdir
from os.path import isfile, join, basename, realpath, isabs, split

from graphene import ObjectType, relay, String, Field
from ml_dash import schema
from ml_dash.schema import files
from ml_dash.schema.files.file_helpers import find_files
from ml_dash.schema.files.metrics import find_metrics
from ml_dash.schema.files.parameters import find_parameters


class Experiment(ObjectType):
    class Meta:
        interfaces = relay.Node,

    name = String(description='name of the directory')
    path = String(description="path to the experiment")

    readme = Field(lambda: schema.files.File)
    parameters = Field(lambda: files.parameters.Parameters, )
    metrics = Field(lambda: files.metrics.Metrics)

    def resolve_readme(self, info, *args, **kwargs):
        # note: keep it simple, just use README for now.
        readmes = schema.files.find_files_by_query(cwd=self.id, query="README.md")
        return readmes[0] if readmes else None

    def resolve_parameters(self, info):
        # note: when called with wrong path, parasitically
        #  slow b/c list all metric files.
        for p in find_parameters(self.id):
            return p
        return None

    def resolve_metrics(self, info):
        # note: when called with wrong path, parasitically
        #  slow b/c list all metric files.
        for m in find_metrics(self.id):
            return m
        return None

    directories = relay.ConnectionField(lambda: schema.directories.DirectoryConnection)
    files = relay.ConnectionField(lambda: schema.files.FileConnection)

    def resolve_directories(self, info, **kwargs):
        from ml_dash.config import Args
        root_dir = join(Args.logdir, self.id[1:])
        return [schema.directories.get_directory(join(self.id, _))
                for _ in listdir(root_dir) if not isfile(join(root_dir, _))]

    def resolve_files(self, info, **kwargs):
        from ml_dash.config import Args
        root_dir = join(Args.logdir, self.id[1:])
        return [schema.files.File(id=join(self.id, _), name=_)
                for _ in listdir(root_dir) if isfile(join(root_dir, _))]

    @classmethod
    def get_node(cls, info, id):
        return get_experiment(id)


class ExperimentConnection(relay.Connection):
    class Meta:
        node = Experiment


def find_experiments(cwd, stop, **kwargs):
    """
    find all experiments

    :param cwd:
    :param start:
    :param stop:
    :return:
    """
    from ml_dash.config import Args
    assert isabs(cwd), "the current work directory need to be an absolute path."
    _cwd = realpath(join(Args.logdir, cwd[1:])).rstrip('/')
    parameter_files = find_files(_cwd, "**/parameters.pkl", stop=stop + 1, **kwargs)
    return [
        # note: not sure about the name.
        Experiment(id=join(cwd.rstrip('/'), p['dir']),
                   name=basename(p['dir']) or ".",
                   path=join(cwd.rstrip('/'), p['dir']),
                   parameters=join(cwd.rstrip('/'), p['path']), )
        for p in parameter_files
    ]


def get_experiment(id):
    _id = id.rstrip('/')
    return Experiment(id=_id, name=split(_id[1:])[-1], path=_id)



import os
import stat
from glob import iglob
from shutil import rmtree
from sanic import response

from . import config


def get_type(mode):
    if stat.S_ISDIR(mode) or stat.S_ISLNK(mode):
        type = 'dir'
    else:
        type = 'file'
    return type


async def remove_path(request, file_path=""):
    print(file_path)
    path = os.path.join(config.Args.logdir, file_path)
    if os.path.isdir(path):
        rmtree(path)
        res = response.text("ok", status=204)
    elif os.path.isfile(path):
        os.remove(path)
        res = response.text("ok", status=204)
    else:
        res = response.text('Not found', status=404)
    return res


from contextlib import contextmanager


@contextmanager
def cwdContext(path):
    owd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(owd)


async def batch_get_path(request):
    try:
        data = request.json

        file_paths = data['paths']
        options = data['options']

        batch_res_data = dict()

        if options.get('json', False):
            for path in file_paths:
                from ml_logger.helpers import load_from_pickle
                batch_res_data[path] = [_ for _ in load_from_pickle(path)]

            res = response.json(batch_res_data, status=200, content_type='application/json')
            return res

    except Exception as e:
        print('Exception: ', e)
        res = response.text('Internal Error' + str(e), status=502)
        return res


async def get_path(request, file_path=""):
    print(file_path)

    as_records = request.args.get('records')
    as_json = request.args.get('json')
    as_log = request.args.get('log')
    as_attachment = int(request.args.get('download', '0'))
    is_recursive = request.args.get('recursive')
    show_hidden = request.args.get('hidden')
    query = request.args.get('query', "*").strip()

    _start = request.args.get('start', None)
    _stop = request.args.get('stop', None)
    start = None if _start is None else int(_start)
    stop = None if _stop is None else int(_stop)

    reservoir_k = int(request.args.get('reservoir', '200'))

    # limit for the search itself.
    search_limit = 500

    path = os.path.join(config.Args.logdir, file_path)
    print("=============>", [query], [path], os.path.isdir(path))

    if os.path.isdir(path):
        from itertools import islice
        with cwdContext(path):
            print(os.getcwd(), query, is_recursive)
            file_paths = list(islice(iglob(query, recursive=is_recursive), start or 0, stop or 200))
            files = map(file_stat, file_paths)
            res = response.json(files, status=200)
    elif os.path.isfile(path):
        if as_records:
            from ml_logger.helpers import load_pickle_as_dataframe
            df = load_pickle_as_dataframe(path, reservoir_k)
            res = response.text(df.to_json(orient="records"), status=200, content_type='application/json')
        elif as_log:
            from ml_logger.helpers import load_pickle_as_dataframe
            df = load_pickle_as_dataframe(path, reservoir_k)
            res = response.text(df.to_json(orient="records"), status=200, content_type='application/json')
        elif as_json:
            from ml_logger.helpers import load_from_pickle
            data = [_ for _ in load_from_pickle(path)]
            res = response.json(data, status=200, content_type='application/json')
        elif type(start) is int or type(stop) is int:
            from itertools import islice
            with open(path, 'r') as f:
                text = ''.join([l for l in islice(f, start, stop)])
            res = response.text(text, status=200)
        else:
            # todo: check the file handling here. Does this use correct
            #  mimeType for text files?
            res = await response.file(path)
            if as_attachment:
                res.headers['Content-Disposition'] = 'attachment'
    else:
        res = response.text('Not found', status=404)
    return res


# use glob! LOL
def file_stat(file_path):
    # this looped over is very slow. Fine for a small list of files though.
    stat_res = os.stat(file_path)
    ft = get_type(stat_res.st_mode)
    sz = stat_res.st_size
    return dict(
        name=os.path.basename(file_path),
        path=file_path,
        mtime=stat_res.st_mtime,
        ctime=stat_res.st_ctime,
        type=ft,
        size=sz,
    )



from os import path
from setuptools import setup, find_packages

cwd = path.dirname(__file__)
with open(path.join(cwd, 'README'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(cwd, 'VERSION'), encoding='utf-8') as f:
    version = f.read()

setup(name="ml-dash",
      description="A Beautiful Visualization Dashboard For Machine Learning",
      long_description=long_description,
      version=version,
      url="https://github.com/episodeyang/ml_logger/tree/master/ml-dash-server",
      author="Ge Yang",
      author_email="yangge1987@gmail.com",
      license=None,
      keywords=["ml_logger",
                "ml-logger",
                "ml dash",
                "ml-dash",
                "ml_dash",
                "dashboard",
                "machine learning",
                "vis_server",
                "logging",
                "debug",
                "debugging"],
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Science/Research",
          "Programming Language :: Python :: 3"
      ],
      packages=[p for p in find_packages() if p != "tests"],
      include_package_data=True,
      install_requires=[
          "cloudpickle",
          'dill',
          'graphene',
          "graphql-core",
          "graphql-relay",
          "graphql-server-core",
          "multidict==4.6.1",
          "numpy",
          'pandas',
          "params_proto",
          "requests",
          "requests_futures",
          'ruamel.yaml',
          'sanic',
          'sanic-cors',
          'Sanic-GraphQL',
          "termcolor",
          "typing"
      ])



from os.path import split
from graphene import ObjectType, relay, String
from ml_dash import schema


class File(ObjectType):
    class Meta:
        interfaces = relay.Node,

    name = String(description='name of the directory')

    # description = String(description='string serialized data')
    # experiments = List(lambda: schema.Experiments)

    @classmethod
    def get_node(cls, info, id):
        return get_file(id)


class FileConnection(relay.Connection):
    class Meta:
        node = File


def get_file(id):
    # path = os.path.join(Args.logdir, id[1:])
    return File(id=id, name=split(id[1:])[1])



from os.path import expanduser

TEST_LOG_DIR = expanduser('~/ml-logger-debug')


def pytest_addoption(parser):
    parser.addoption('--log-dir',
                     action='store',
                     default=TEST_LOG_DIR,
                     help="The logging path for the test.")



from os.path import isfile
from graphene import ObjectType, relay, String
from ml_dash import schema


class User(ObjectType):
    class Meta:
        interfaces = relay.Node,

    @classmethod
    def get_node(_, info, id):
        print(info, id)
        return get_user(id)

    username = String(description='string serialized data')
    name = String(description='string serialized data')

    projects = relay.ConnectionField(lambda: schema.projects.ProjectConnection)

    def resolve_projects(self, info, before=None, after=None, first=None, last=None):
        # todo: figure out a good way for pagination.
        # note: project does not support first, last
        return schema.projects.get_projects(self.username)

    # teams = List(lambda: schema.Team)


def get_users(ids=None):
    import os
    from ml_dash.config import Args
    return [User(username=_, name="Ge Yang") for _ in os.listdir(Args.logdir) if not isfile(_)]


def get_user(username):
    return User(username=username, name="Ge Yang", id=username)
